
%% dining philosophers


:-object phil.

var name, room, food=4, lfork, rfork.

phil(Name, Room, Lf, Rf) :-
	name := Name,
	room := Room,
	lfork := Lf,
	rfork := Rf,
	proceed(Room).

thinking :-
	format('~w:~t~w thinking ... (food=~w)~n', [this, name, food]).
eating :-
	format('~w:~t~w eating ..... (food=~w)~n', [this, name, food]),
	-- food.

proceed(Room) :-
	thinking,
	( food > 0 ->
		Room <- enter,
		lfork <- pickup(name),
		rfork <- pickup(name),
		eating,
		lfork <- putdown(name),
		rfork <- putdown(name),
		Room <- leave,
		proceed(Room)
	;
		lfork <- dont_need_fork(name),
		rfork <- dont_need_fork(name),
		Room <- going_home(name)
	).

:-end_object(phil).


:-object(room).

var nphils = 0, mphils = 0, occupancy = 0.

enter :- ++ occupancy.	%% occupancy := occupancy + 1.
leave :- -- occupancy.	%% occupancy := occupancy - 1.

room(P) :-
	nphils := P,
	mphils := P - 1,
	guard.

guard :-
	( nphils > 0 ->
		accept(
			enter <== [occupancy < mphils],
			leave <== [occupancy > 0],
			going_home(_)
		),
		guard
	;
		format('~nno philosophers left~n', [])
	).

going_home(Phil) :-
	nphils := nphils - 1,
	format('~w:~t~w going home (~w philosophers left)~n', [this, Phil, nphils]).

:-end_object(room).


:-object(fork).

var label, used=false, need=2.

pickup(Phil) :-
	used := true,
	format('~w:~t~w pickup (~w)~n', [this, label, Phil]).
putdown(Phil) :-
	used := false,
	format('~w:~t~w putdown (~w)~n', [this, label, Phil]).

dont_need_fork(Phil) :-
	format('~w:~t~w dont_need_fork (~w)~n', [this, label, Phil]),
	need := need - 1.

fork(N,L) :-
	need := N,
	label := L,
	state.

state :-
	( need > 0 ->
		accept(
			pickup(_) <== [used = false],
			putdown(_) <== [used = true],
			dont_need_fork(_)
		),
		state
	;
		format('~w~texit ~w~n', [this, label])
	).

:-end_object(fork).


:-object pxphil.

main :-
	F1 := new(fork(2, f1)),
	F2 := new(fork(2, f2)),
	F3 := new(fork(2, f3)),
	F4 := new(fork(2, f4)),
	F5 := new(fork(2, f5)),

	DR := new(room(5)),

	P1 := new(phil(p1,DR,F1,F2)),
	P2 := new(phil(p2,DR,F2,F3)),
	P3 := new(phil(p3,DR,F3,F4)),
	P4 := new(phil(p4,DR,F4,F5)),
	P5 := new(phil(p5,DR,F5,F1)).

:-end_object pxphil.

